from dotenv import load_dotenv

from langchain.chains import LLMChain
from langchain_core.prompts import PromptTemplate

from rqle_ai_langchain_util import settings
from rqle_ai_langchain_util.utils import video_util
from rqle_ai_langchain_util.llms.adapters.llm_adapters import LLMAdapter
from rqle_ai_langchain_util.llms.llm_mediator import LLMMediator

# logging configuration
from time import time
import logging
from logging.config import dictConfig
dictConfig(settings.LOG_CONFIG)
logger = logging.getLogger(__name__)

load_dotenv()


class PodProtocol:

    def info(self):
        """
        :return: the name of the class
        """
        return self.__class__.__name__

    def __init__(self, config_folder: str):
        """
        :param config_folder: the folder containing the configuration for execution
        """
        self.config_folder = config_folder
        self.llm_mediator = LLMMediator(LLMAdapter.OLLAMA_AI, self.config_folder)

    def load_chain(self):
        try:
            # configure the prompt
            prompt = PromptTemplate(template=self.llm_mediator.prompt_template.prompt,
                                    input_variables=['user_input'])
            logger.debug(f'Generated prompt: {prompt}')

            # return the LLM
            return LLMChain(llm=self.llm_mediator.model, prompt=prompt)
        except Exception as e:
            logger.error(f'Error loading chain: {self.config_folder}\n{e}', exc_info=True)
            raise e

    def invoke_chain(self, video_file: str):
        try:
            # retrieve the text transcription of the video
            audio_filename = video_util.video_to_wav(video_file)
            transcribed_text = video_util.transcribe_audio(audio_filename, speech_recognition_engine='vosk')

            # invoke the chain on the transcribed text
            chain = self.load_chain()
            output = chain.invoke({'user_input': transcribed_text})
            logger.debug(f'Output from PodProtocol: {output}')
            return output
        except Exception as e:
            logger.error(f'Error executing chain: {self.config_folder}\n{e}', exc_info=True)


if __name__ == '__main__':
    start = time()
    try:
        pod_protocol = PodProtocol(config_folder='pod_protocol')
        pod_protocol.invoke_chain(video_file='C:/Users/quent/Videos/Podcasts/TalkLab_interviews/EP22_AI_Episode-3.2.mp4')
    except Exception:
        pass
    finally:
        # calculate elapsed time
        total_time = round((time() - start) / 60, 3)
        logger.info(f'Elapsed time to generate summary: {total_time} min')



from dotenv import load_dotenv

from langchain_openai import AzureOpenAI, AzureChatOpenAI, AzureOpenAIEmbeddings

from rqle_ai_langchain_util.prompts.prompt_config import PromptConfig, PromptTypeEnum

load_dotenv()


def _load_azure_openai_llm_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Azure OpenAI LLMs
    """
    llm = AzureOpenAI(
        deployment_name=config.model_name,
        model_name=config.model_name,
        temperature=config.parameters.temperature,
        max_tokens=config.parameters.max_tokens,
        top_p=config.parameters.top_p,
        presence_penalty=config.parameters.presence_penalty,
        frequency_penalty=config.parameters.frequency_penalty
    )
    return llm


def _load_azure_openai_chat_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Azure OpenAI LLMs
    """
    llm = AzureChatOpenAI(
        deployment_name=config.model_name,
        model_name=config.model_name,
        temperature=config.parameters.temperature,
        max_tokens=config.parameters.max_tokens
    )
    return llm


def _load_azure_openai_embeddings_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Azure OpenAI LLMs
    """
    llm = AzureOpenAIEmbeddings(model=config.model_name)
    return llm


def load_azure_openai_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Azure OpenAI LLMs
    """
    if config.type == PromptTypeEnum.chat:
        return _load_azure_openai_chat_from_prompt_config(config)
    elif config.type == PromptTypeEnum.completion:
        return _load_azure_openai_llm_from_prompt_config(config)
    elif config.type == PromptTypeEnum.embedding:
        return _load_azure_openai_embeddings_from_prompt_config(config)



from enum import Enum


class LLMAdapter(str, Enum):
    AWS_BEDROCK = "aws_bedrock_adapter"
    AZURE_OPENAI = "azure_openai_adapter"
    GOOGLE_GEMINI = "google_gemini_adapter"
    OLLAMA_AI = "ollama_adapter"
    OCI_AI = "oci_ai_adapter"

    __all__ = [
        AWS_BEDROCK,
        AZURE_OPENAI,
        GOOGLE_GEMINI,
        OLLAMA_AI,
        OCI_AI
    ]



from typing import List
from pydantic import BaseModel, Field

from rqle_ai_langchain_util.utils.file_util import read_file


class PromptExample(BaseModel):
    version: int = Field(description='Schema version', default=1)
    examples: List[dict] = Field(description='Examples of the prompt', default_factory=list)

    @staticmethod
    def from_json(example_json_path: str) -> 'PromptExample':
        """
        Create a PromptConfig from a JSON string
        :param example_json_path: Prompt configuration as a JSON string
        :return: a PromptConfig object
        """
        return PromptExample.model_validate_json(read_file(example_json_path, 'example.json'))


# Information about the different providers supported by AWS Bedrock -
# https://us-west-2.console.aws.amazon.com/bedrock/home
import os
from dotenv import load_dotenv

from langchain_aws.llms.bedrock import BedrockLLM
from langchain_aws.chat_models.bedrock import ChatBedrock
from langchain_aws.embeddings.bedrock import BedrockEmbeddings

from rqle_ai_langchain_util.prompts.prompt_config import PromptConfig, PromptTypeEnum

load_dotenv()


def _get_model_kwargs(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: the model arguments relevant for different models supported in AWS bedrock
    """
    provider = config.model_name.split('.')[0]
    if provider == 'ai21':
        return {
            "maxTokens": config.parameters.max_tokens,
            "temperature": config.parameters.temperature,
            "topP": config.parameters.top_p,
            "presencePenalty":{"scale": config.parameters.presence_penalty},
            "frequencyPenalty":{"scale": config.parameters.frequency_penalty}
        }
    elif provider == 'amazon':
        # TODO investigate '/basic-syntax/' text when using completion
        return {
            "maxTokenCount": config.parameters.max_tokens,
            "temperature": config.parameters.temperature,
            "topP": config.parameters.top_p
        }
    elif provider == 'cohere':
        return {
            "max_tokens": config.parameters.max_tokens,
            "temperature": config.parameters.temperature,
            "p": config.parameters.top_p
        }
    elif provider == 'meta':
        # TODO investigate why parameters work in direct call, but not when passed through UI
        # Malformed input request: #: extraneous key [max_gen_length] is not permitted
        return {
            "max_gen_len": config.parameters.max_tokens,
            "temperature": config.parameters.temperature,
            "top_p": config.parameters.top_p
        }
    elif provider == 'mistral':
        # TODO investigating issue with mistral model
        # Malformed input request: #: required key [prompt] not found#: extraneous key [inputText] is not permitted
        return {
            "max_tokens": config.parameters.max_tokens,
            "temperature": config.parameters.temperature,
            "top_p": config.parameters.top_p
        }
    else:
        raise NotImplementedError(f'LLM provider {provider} not supported')


def _load_aws_bedrock_llm_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for AWS Bedrock LLMs
    """
    # transform the model configuration in the appropriate dictionary
    model_kwargs = _get_model_kwargs(config)
    # create a new Bedrock LLM object
    llm = BedrockLLM(
        credentials_profile_name=os.environ.get("BWB_PROFILE_NAME"),
        region_name=os.environ.get("BWB_REGION_NAME"),
        endpoint_url=os.environ.get("BWB_ENDPOINT_URL"),
        model_id=config.model_name,
        model_kwargs= model_kwargs
    )
    return llm


def _load_aws_bedrock_chat_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain Chat object configured for AWS Bedrock LLMs
    """
    # transform the model configuration in the appropriate dictionary
    model_kwargs = _get_model_kwargs(config)
    # create a new Bedrock LLM object
    llm = ChatBedrock(
        credentials_profile_name=os.environ.get("BWB_PROFILE_NAME"),
        region_name=os.environ.get("BWB_REGION_NAME"),
        endpoint_url=os.environ.get("BWB_ENDPOINT_URL"),
        model_id=config.model_name,
        model_kwargs=model_kwargs
    )
    return llm


def _load_aws_bedrock_embeddings_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain embeddings object configured for AWS Bedrock LLMs
    """
    llm = BedrockEmbeddings(
        credentials_profile_name=os.environ.get("BWB_PROFILE_NAME"),
        region_name=os.environ.get("BWB_REGION_NAME"),
        endpoint_url=os.environ.get("BWB_ENDPOINT_URL"),
        model_id=config.model_name
    )
    return llm


def load_aws_bedrock_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for AWS Bedrock LLMs
    """
    if config.type == PromptTypeEnum.chat:
        return _load_aws_bedrock_chat_from_prompt_config(config)
    elif config.type == PromptTypeEnum.completion:
        return _load_aws_bedrock_llm_from_prompt_config(config)
    elif config.type == PromptTypeEnum.embedding:
        return _load_aws_bedrock_embeddings_from_prompt_config(config)



import os
from dotenv import load_dotenv

from langchain.callbacks.manager import CallbackManager
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain_community.llms.ollama import Ollama
from langchain_community.chat_models import ChatOllama
from langchain_community.embeddings import OllamaEmbeddings

from rqle_ai_langchain_util.prompts.prompt_config import PromptConfig, PromptTypeEnum

load_dotenv()


def _load_ollama_llm_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a Langchain object configured for Ollama LLMs
    """
    llm = Ollama(
        base_url=os.getenv('OLLAMA_ENDPOINT'),
        model=config.model_name,
        temperature=config.parameters.temperature,
        callbacks=CallbackManager([StreamingStdOutCallbackHandler()])
    )
    return llm


def _load_ollama_chat_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a Langchain chat object configured for Ollama LLMs
    """
    llm = ChatOllama(
        base_url=os.getenv('OLLAMA_ENDPOINT'),
        model=config.model_name,
        temperature=config.parameters.temperature,
        callbacks=CallbackManager([StreamingStdOutCallbackHandler()])
    )
    return llm


def _load_ollama_embeddings_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the embeddings execution
    :return: a Langchain object configured for Ollama LLMs
    """
    llm = OllamaEmbeddings(
        base_url=os.getenv('OLLAMA_ENDPOINT'),
        model=config.model_name
    )
    return llm


def load_ollama_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a OpenAI object configured for a local LLM
    """
    if config.type == PromptTypeEnum.chat:
        return _load_ollama_chat_from_prompt_config(config)
    elif config.type == PromptTypeEnum.completion:
        return _load_ollama_llm_from_prompt_config(config)
    elif config.type == PromptTypeEnum.embedding:
        return _load_ollama_embeddings_from_prompt_config(config)



import streamlit as st
from streamlit_tags import st_tags
from PIL import Image

from rqle_ai_langchain_util import settings
from rqle_ai_langchain_util.utils import gui_util
from rqle_ai_langchain_util.llms import llm_guard_validator
from rqle_ai_langchain_util.apps.techtype_rocket.techtype_rocket import TechTypeRocket

import logging
from logging.config import dictConfig
dictConfig(settings.LOG_CONFIG)
logger = logging.getLogger(__name__)

APPLICATION_LOGO_PATH = 'img/techtype_rocket_logo.jpg'
APPLICATION_LOGO = Image.open(APPLICATION_LOGO_PATH)
APPLICATION_NAME = 'TechType Rocket'
APPLICATION_DESCRIPTION = f"""
    **{APPLICATION_NAME}**:copyright: is your one-stop platform designed to equip aspiring writers with the skills and 
    guidance to confidently create engaging tech blogs for diverse audiences.
"""

st.set_page_config(page_title=APPLICATION_NAME, page_icon=APPLICATION_LOGO,
                   layout='wide', initial_sidebar_state='collapsed')

# information about / help component of the application
st.sidebar = gui_util.set_page_sidebar(application_description=APPLICATION_DESCRIPTION)

# set the logo / title of the application
st.markdown(body=gui_util.set_page_header(application_name=APPLICATION_NAME,
                                           application_logo_path=APPLICATION_LOGO_PATH,
                                           background_color='#334246',
                                           text_color='#FFFFFF'),
            unsafe_allow_html=True)

# main body of the application
input, output, analytics = st.columns([1.5, 2, 0.5])
output.subheader('Generated Blog', divider='rainbow')
analytics.subheader('Analytics', divider='rainbow')

with ((input)):
    with st.form(key='blog_input'):
        st.subheader('Enter Blog Details', divider='rainbow')
        target_reading_time = st.number_input(label='Blog length (min)', min_value=5, step=1,
                                              help='Enter how long you want blog to be in terms of reading time')
        target_audience = st.selectbox(label='Target Audience', options=('Technology Leaders', 'Business Leaders', 'Novice'))
        # determine the focus of the blog
        topic = st.text_area(label='Topic', height=200)
        keywords = st_tags(label='Keywords', text='Press enter to add more keywords', maxtags=10)
        # determine whether the guard rails should be applied
        guard_rails_enabled = st.checkbox(label='Enable Guard Rails', value=True)
        # form submission button
        submit = st.form_submit_button(label='Generate', type='primary')

        # execute logic if submit button is clicked
        if submit:
            logger.debug(f'User input: target_reading_time={target_reading_time}, target_audience={target_audience}, topic={topic}, keywords={keywords}')
            # Checked that the required fields are not empty
            if (not topic and topic != '') or len(keywords) == 0:
                st.error('Please enter a few sentence describing the topic and at least one keyword to be written about.')
                logger.error('Error: No topics and / or keywords were entered by user.')
            else:
                with output:
                    with st.spinner('Generating blog...'):
                        try:
                            # check whether guard rails should be checked on the input
                            if guard_rails_enabled:
                                input_is_valid, input_error_msg = llm_guard_validator.validate_llm_input(input=topic)
                                # raise an exception if any of the guard rails are violated
                                if not input_is_valid:
                                    raise IOError(input_error_msg)
                            # execute the LLM to generate the blog
                            tech_type_rocket = TechTypeRocket(config_folder='techtype_rocket')
                            generated_blog = tech_type_rocket.invoke_chain(target_reading_time=target_reading_time,
                                                                           target_audience=target_audience,
                                                                           topic=topic,
                                                                           keywords=keywords)
                            generated_blog_data = generated_blog['text']
                            # check whether guard rails should be checked on the output
                            if guard_rails_enabled:
                                output_is_valid, output_error_msg = llm_guard_validator.validate_llm_output(prompt=f'Generate a blog post about "{topic}" targeting {target_audience} '
                                                                                                                   f'with an estimated reading time of {target_reading_time} minutes.',
                                                                                                            output=generated_blog_data)
                                if not output_is_valid:
                                    raise IOError(output_error_msg)
                            markdown_tab, text_tab = st.tabs(['WYSIWYG', 'Markdown'])
                            with markdown_tab:
                                st.markdown(generated_blog_data)
                            with text_tab:
                                st.text(generated_blog_data)
                            # analytics component
                            with analytics:
                                st.subheader('Reading Time')
                                estimated_reading_time = llm_guard_validator.estimate_reading_time(generated_blog_data)
                                logger.debug(f'Estimated reading time of generated blog is {estimated_reading_time} minutes,'
                                             f'while the target was {target_reading_time} minutes.')
                                chart = gui_util.make_donut_graph(expected=target_reading_time, actual=estimated_reading_time)
                                st.plotly_chart(chart, use_container_width=True)
                        except Exception as e:
                            st.error(f'An error occurred while generating the blog: {e}')
                            logger.error(f'Error occurred while generating the blog: {e}')

# information about the footer of the application (including logo and years)
st.markdown(body=gui_util.set_page_footer(), unsafe_allow_html=True)


import os
from dotenv import load_dotenv

from langchain_community.llms import OCIGenAI
from langchain_community.embeddings import OCIGenAIEmbeddings
from rqle_ai_langchain_util.prompts.prompt_config import PromptConfig, PromptTypeEnum

load_dotenv()


def _load_oci_ai_llm_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Oracle Cloud AI LLMs
    """
    llm = OCIGenAI(
        model_id=config.model_name,
        service_endpoint=os.getenv('OCI_AI_ENDPOINT'),
        compartment_id=os.getenv('OCI_AI_COMPARTMENT_ID'),
        model_kwargs={
            'max_tokens': config.parameters.max_tokens,
            'temperature': config.parameters.temperature,
            'top_p': config.parameters.top_p,
        }
    )
    return llm


def _load_oci_ai_embeddings_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Oracle Cloud AI LLMs
    """
    llm = OCIGenAIEmbeddings(
        model_id=config.model_name,
        service_endpoint=os.getenv('OCI_AI_ENDPOINT'),
        compartment_id=os.getenv('OCI_AI_COMPARTMENT_ID')
    )
    return llm


def load_oci_ai_llm_from_prompt_config(config: PromptConfig):
    """
    :param config: the configuration for the LLM execution
    :return: a LangChain object configured for Oracle Cloud AI LLMs
    """
    if config.type == PromptTypeEnum.completion:
        return _load_oci_ai_llm_from_prompt_config(config)
    elif config.type == PromptTypeEnum.embedding:
        return _load_oci_ai_embeddings_from_prompt_config(config)
    else:
        raise NotImplementedError(f'LLM type {config.type} not supported')



from dotenv import load_dotenv

from langchain.chains import LLMChain
from langchain_core.prompts import PromptTemplate

from rqle_ai_langchain_util import settings
from rqle_ai_langchain_util.llms.adapters.llm_adapters import LLMAdapter
from rqle_ai_langchain_util.llms.llm_mediator import LLMMediator

# logging configuration
from time import time
import logging
from logging.config import dictConfig
dictConfig(settings.LOG_CONFIG)
logger = logging.getLogger(__name__)

load_dotenv()


class TechTypeRocket:

    def info(self):
        """
        :return: the name of the class
        """
        return self.__class__.__name__

    def __init__(self, config_folder: str):
        # configure the LLM to be executed
        self.config_folder = config_folder
        self.llm_mediator = LLMMediator(LLMAdapter.OLLAMA_AI, self.config_folder)

    def load_chain(self):
        """
        :return: a configured LLM to be executed
        """
        try:
            # configure the prompt
            prompt = PromptTemplate(template=self.llm_mediator.prompt_template.prompt,
                                    input_variables=['target_reading_time', 'target_audience', 'topic', 'keywords'])
            logger.debug(f'Generated prompt: {prompt}')

            # return the LLM
            return LLMChain(llm=self.llm_mediator.model, prompt=prompt)
        except Exception as e:
            logger.error(f'Error loading chain: {self.config_folder}\n{e}', exc_info=True)
            raise e

    def invoke_chain(self, target_reading_time: int, target_audience: str, topic: str, keywords: list[str]):
        """
        :param target_reading_time: the targeted length of the blog in terms of reading time
        :param target_audience: the target audience of the blog
        :param topic: the topic of the blog
        :param keywords: additional keywords related to the topic
        :return: the output from executing the LLM chain
        """
        try:
            # load the chain to be executed
            chain = self.load_chain()
            # execute the chain
            output = chain.invoke({'target_reading_time': target_reading_time,
                                   'target_audience': target_audience,
                                   'topic': topic,
                                   'keywords': keywords})
            logger.debug(f'Output from TechType Rocket: {output}')
            return output
        except Exception as e:
            logger.error(f'Error executing chain: {self.config_folder}\n{e}', exc_info=True)
            raise e


if __name__ == '__main__':
    start = time()
    try:
        # create an instance of the class
        tech_type_rocket = TechTypeRocket(config_folder='techtype_rocket')
        # invoke the chain
        output = tech_type_rocket.invoke_chain(target_reading_time=5,
                                               target_audience='Business Leaders',
                                               topic='Discuss how generative AI is democratizing the adoption of AI,'
                                                     'Include a discussion on how generative AI removes barriers '
                                                     'provided by discriminative AI',
                                               keywords=['generative AI', 'discriminative AI', 'AI adoption'])
    except Exception as e:
        logger.error(f'Error executing TechType Rocket\n{e}', exc_info=True)
    finally:
        # calculate elapsed time
        total_time = round((time() - start) / 60, 3)
        logger.info(f'Elapsed time to execute TechType Rocket: {total_time} min')